/***** spin: run.c *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include <stdlib.h>
#include "spin.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif

extern RunList	*X, *run;
extern Symbol	*Fname;
extern Element	*LastStep;
extern int	Rvous, lineno, Tval, interactive, MadeChoice;
extern int	TstOnly, verbose, s_trail, xspin, jumpsteps, depth;
extern int	nproc, nstop, no_print;
extern int      Expand_Ok;

extern Symbol	*has_lab(Element *);
extern Symbol   *struct_sym(Lextok *, Symbol *);
extern Lextok   *struct_tok(Lextok *, Symbol *);

static long	Seed = 1;
static int	E_Check = 0;

static int	eval_sync(Element *);
static int	interprint(Lextok *);
static int	pc_enabled(Lextok *n);

Value zero;


void
Srand(unsigned int s)
{	Seed = s;
}

long
Rand(void)
{	/* CACM 31(10), Oct 1988 */
	Seed = 16807*(Seed%127773) - 2836*(Seed/127773);
	if (Seed <= 0) Seed += 2147483647;
	return Seed;
}


static int
push_symbol(RunList *rl, Symbol *s)
{
  int old_sp = rl->stk->used;

  if (s->type == FTYPE) {
    s->Sval = (Lextok **)stack_push(rl->stk, var_size(s) + sizeof(Lextok *));
    s->Sval[0] = (Lextok *)s->nel;
  }
  else {
    s->val = (int *)stack_push(rl->stk, var_size(s) + sizeof(int));
    s->val[0] = s->nel;
  }

  return rl->stk->used - old_sp;
}


static int
push_struct(RunList *rl, Symbol *s)
{
  Lextok *fp, *tl;
  int i, old_sp = rl->stk->used;

  if (s->type != STRUCT)
    push_symbol(rl, s);

  s->Slst = s->Slst;
  s->Sval = (Lextok **) stack_push(rl->stk, (s->nel + 1) * sizeof(Lextok *));  
  s->Sval[0] = (Lextok *) s->nel;

  for (i = 0; i < s->nel; i++)
    {	
      s->Sval[i + 1] = cpnn(s->Slst, 1, 1, 1);
      for (fp = s->Sval[i + 1]; fp; fp = fp->rgt)
	for (tl = fp->lft; tl; tl = tl->rgt)
	  push_struct(rl, tl->sym);
    }
  
  return rl->stk->used - old_sp;
}


static Symbol *
dup_symbol(Sequence *sq, Symbol *s)
{
  Symbol *t = (Symbol *) emalloc(sizeof(Symbol));

  memcpy(t, s, sizeof(Symbol)); 
  t->next = sq->symtab;   /* add it */
  sq->symtab = t; 
  
  return t;
}


void
rt_enter(RunList *rl, Sequence *s, int tag)
{      
  SeqList *t = (SeqList *) emalloc(sizeof(SeqList));
  Sequence *sq = (Sequence *) emalloc(sizeof(Sequence));
  Symbol *stab, *v;
  int frame = 0;
  
  sq->cnxt    = s->cnxt;
  sq->frst    = s->frst;
  sq->last    = s->last;
  sq->extent  = s->extent;
  sq->maxel   = s->maxel;

  t->this     = sq;
  t->nxt      = rl->rt_seq;
  rl->rt_seq  = t;  

  for (stab = s->symtab; stab; stab = stab->next)
    {
      v = dup_symbol(rl->rt_seq->this, stab);
      /* Don't allocate space for pointers */
      if (!(v->type & POINTER) && v->type != LABEL)
	{
	  if (v->type != STRUCT)
	    frame += push_symbol(rl, v);
	  else
	    frame += push_struct(rl, v);
	}
    }

  /* Push the stack frame size */
  *(int *)stack_push(rl->stk, sizeof(int)) = frame;

  /* Push the stack frame tag */
  *(int *)stack_push(rl->stk, sizeof(int)) = tag;
}


int
rt_exit(RunList *rl, int tag)
{
  Symbol *stab, *snxt;
  SeqList *nxt;
  int frame;
  int stag;

  for (;;)
    {
      /* Peek stack frame tag */
      stag = *(int *)stack_peek(rl->stk, sizeof(int));
      stack_pop(rl->stk, sizeof(int));
	
      /* Peek stack frame size */
      frame = *(int *)stack_peek(rl->stk, sizeof(int));
      stack_pop(rl->stk, sizeof(int) + frame);

      for (stab = rl->rt_seq->this->symtab; stab; stab = snxt)
	{
	  snxt = stab->next;
	  free(stab);
	}

      nxt = rl->rt_seq->nxt;
      free(rl->rt_seq);
  
      rl->rt_seq = nxt;

      if (stag == tag)
	return 1;

      switch (stag) 
	{
	case FUNCALL_TAG:
	  /* Pop return symbol and return address */
	  stack_pop(X->stk, sizeof(Lextok *) + sizeof(Element *));
	  break;

	default:
	  return 0;
	}
    }
}


Element *
eval_funret(Value *vp, int rtype)
{
  char *sp;
  int frame;
  Element *e;
  int *val;
  Lextok *n;

  rt_exit(X, FUNCALL_TAG);

  /* Pop the return symbol (kinna ugly !) */
  n = *(Lextok **)stack_peek(X->stk, sizeof(Lextok *));
  stack_pop(X->stk, sizeof(Lextok *));

  /* Set the return value */
  if (n)
    assign(n, vp, rtype);

  /* Peek the return address */
  e = *(Element **)stack_peek(X->stk, sizeof(Element *));
  
  /* Pop the return address */
  stack_pop(X->stk, sizeof(Element *));
  return e;
}


void
block_exit(RunList *X, SeqList *sl)
{
  if (!sl)
    return;

  block_exit(X, sl->nxt);
  rt_exit(X, LOCAL_TAG);
}


Element *
eval_sub(Element *e)
{	Element *f, *g, *next;
	SeqList *z;
	int i, j, k;
	Value v;

	if (!e->n)
	  return ZE;

#if (DEBUG >= 1)
	printf("eval_sub(%d %s) ",
		e->Seqno, e->esc?"+esc":"");
	comment(stdout, e->n, 0);
	printf("\n");
#endif

	switch (e->n->ntyp)
	  {
	  case GOTO:
	    if (e->tag & JMP)
	      block_exit(X, e->sel);

	    if (Rvous) 
	      return ZE;
	    
	    f = get_lab(e->n, 1);
	    cross_dsteps(e->n, f->n);
	    return f;
       
	  case UNLESS:
	    /* escapes were distributed into sequence */
	    return eval_sub(e->sub->this->frst);

	  case IF: 
	  case DO:
	    if (e->sub)	/* true for IF, DO, and UNLESS */
	      {	
		Element *has_else = ZE;
		int nr_else, nr_choices = 0;

		if (interactive
		    && !MadeChoice && !E_Check
		    && depth >= jumpsteps)
		  {	printf("Select stmnt (");
			whoruns(0); printf(")\n");
			if (nproc-nstop > 1)
			printf("\tchoice 0: other process\n");
		  }

		for (z = e->sub, j=0; z; z = z->nxt)
		  {	j++;
			if (interactive
			&& !MadeChoice && !E_Check
			&& depth >= jumpsteps
			&& z->this->frst
			&& (xspin || (verbose&32) || Enabled0(z->this->frst)))
			{	if (z->this->frst->n->ntyp == ELSE)
			        {    has_else = (Rvous)?ZE:z->this->frst->nxt;
				     nr_else = j;
				     continue;
				}
				printf("\tchoice %d: ", j);
				if (!Enabled0(z->this->frst))
				  printf("unexecutable, ");
				else
				  nr_choices++;
				comment(stdout, z->this->frst->n, 0);
				printf("\n");
			}	
		  }
		
		if (nr_choices == 0 && has_else)
		  printf("\tchoice %d: (else)\n", nr_else);

		if (interactive && depth >= jumpsteps && !E_Check)
		  {	if (!MadeChoice)
		        {	char buf[256];
				if (xspin)
					printf("Make Selection %d\n\n", j);
				else
					printf("Select [0-%d]: ", j);
				fflush(stdout);
				scanf("%s", buf);
				if (isdigit(buf[0]))
					k = atoi(buf);
				else
				{	if (buf[0] == 'q')
						alldone(0);
					k = -1;
				}
			} else
			{	k = MadeChoice;
				MadeChoice = 0;
			}
			if (k < 1 || k > j)
			{	if (k != 0) printf("\tchoice outside range\n");
				return ZE;
			}
			k--;
		} else
		{	if (e->n && e->n->indstep >= 0)
				k = 0;	/* select 1st executable guard */
			else
				k = Rand()%j;	/* nondeterminism */
		}
		has_else = ZE;
		for (i = 0, z = e->sub; i < j+k; i++)
		{	if (z->this->frst
			&&  z->this->frst->n->ntyp == ELSE)
			{	has_else = (Rvous)?ZE:z->this->frst->nxt;
				if (!interactive || depth < jumpsteps)
				{	z = (z->nxt)?z->nxt:e->sub;
					continue;
			}	}
			if (i >= k)
			{	if (f = eval_sub(z->this->frst))
					return f;
				else if (interactive && depth >= jumpsteps)
				{	if (!E_Check)
						printf("\tunexecutable\n");
					return ZE;
			}	}
			z = (z->nxt)?z->nxt:e->sub;
		}
		LastStep = z->this->frst;
		return has_else;
	      } 

	  case ATOMIC:
	  case D_STEP:
	    f = e->n->sl->this->frst;
	    g = e->n->sl->this->last;	    
	    g->nxt = e->nxt;

	    if (!(g = eval_sub(f)))	/* atomic guard */
	      return ZE;
	    
	    return g;

	  case NON_ATOMIC:
	    f = e->n->sl->this->frst;
	    g = e->n->sl->this->last;	    
	    g->nxt = e->nxt;		/* close it */

	    rt_enter(X, e->n->sl->this, LOCAL_TAG);
	    initlocals(X->rt_seq->this);

	    return eval_sub(f);

	  case '.':   
	    if (!Rvous) return e->nxt;
	    return eval_sub(e->nxt);

	  default:	
	    {  
	      SeqList *x;

	      if (!(e->status & D_ATOM) && e->esc && verbose&32)
		{	
		  printf("Stmnt [");
		  comment(stdout, e->n, 0);
		  printf("] has escape(s): ");
		  for (x = e->esc; x; x = x->nxt)
		    {	
		      printf("[");
		      g = x->this->frst;
		      if (g->n->ntyp == ATOMIC
			  ||  g->n->ntyp == NON_ATOMIC)
			g = g->n->sl->this->frst;
		      comment(stdout, g->n, 0);
		      printf("] ");
		    }
		  printf("\n");
		}
			
	      for (x = e->esc; x; x = x->nxt)
		{	
		  if (g = eval_sub(x->this->frst))
		    {	
		      if (verbose&4)
			printf("\tEscape taken\n");
		      return g;
		    }
		}
	      
	      switch (e->n->ntyp) 
		{
		case TIMEOUT: case RUN: 
		case PRINT: case ASGN: case ASSERT:
		case 's': case 'r': case 'c':
		  /* toplevel statements only */
		  LastStep = e;
		default:
		  break;
		}
			  
	      if (Rvous)
		{
		  /* allow only synchronous receives
		     and related node types    */
		  Lextok *now = (e)?e->n:ZN;
		  
		  if (!now
		      ||  now->ntyp != 'r'
		      ||  now->val >= 2	/* no rv with a poll */
		      ||  !q_is_sync(now))
		    {
		      return 0;
		    }
		  
		  LastStep = e;
		}
	      
	      next = eval_expr(e, e->n, &v);

	      if (e->tag & FE       
		  && e->n->ntyp != FUNCALL
		  && e->n->ntyp != RETFUN
		  && e->n->ntyp != RETURN)
		return eval_funret(0, 0);

	      if (e->tag & BE)
		block_exit(X, e->sel);

	      return v.value ? next : ZE;
	    }
	  }
	
	return ZE; /* not reached */
}


int
node_type(Lextok *now)
{
  switch (now->ntyp) 
    {
    case FULL:	case NFULL:
    case EMPTY:	case NEMPTY:
    case RUN:	case LEN:
      return BYTE;
      
    case LVAL:
      return POINTER | now->lft->sym->type;
         
    case __NULL:
      return POINTER;
      
    default:
      if (now->ntyp == NAME)
	{
	  Symbol *s, *t;
	  Lextok *m;

	  if (!now->sym)
	    fatal("no symbol - node_type", (char *)0);

	  if (!(s = findloc(now->sym)))
	    s = now->sym;

	  if (!s->type)
	    fatal("undeclared variable '%s'\n", s->name);

	  m = struct_tok(now, s);
	  t = struct_sym(now, s);

	  if ((t->type & POINTER) && (t->nel == 1) && m->lft)
	    return t->type & ~POINTER;

	  return t->type;
	}
      
      if (now->ntyp == FNAME)
	return FTYPE;	
 
      return INT;
    }
}


int
assign(Lextok *lft, Value *vp, int rtype)
{	
  int ltype;
  int rval;
  
  if (TstOnly) return 1;

  ltype = node_type(lft);

  if (!typ_ck(ltype, rtype, "assignment"))
    return 1;

  if (ltype == FTYPE)
    return setfunc(lft, (Sequence *)vp->value);
  
  if ((ltype & POINTER) && (rtype & POINTER))
    return setlval(lft, vp->lval, vp->heap);
  
  return setval(lft, vp->value);
}


static int
eval_new(Lextok * now)
{
  Symbol *s = (Symbol *) emalloc(sizeof(Symbol));
  memcpy(s, now->sym, sizeof(Symbol));
  
  if (now->lft)
    s->nel = eval(now->lft);
  else
    s->nel = 1;

  if (s->type == STRUCT) {
    ini_struct(s);
    return (int)s->Sval;
  } else {
    ini_var(s);
    return (int)s->val;
  }
}


static int
eval_delete(Lextok * now)
{
  Symbol *s = now->lft->sym->context ? findloc(now->lft->sym) : now->lft->sym;

  if (!s) 
    fatal("no symbol (delete)", 0);

  if (!(s->type & POINTER)) {
    non_fatal("attempt to delete non pointer '%s'", s->name);
    return 1;
  }

  if (!s->hidden & 64) {
    non_fatal("pointer '%s' not on the heap", s->name);
    return 1;
  }

  if (s->type & STRUCT) 
    {
      if (!s->Sval) {
	non_fatal("pointer '%s' not initialized", s->name);
	return 1;
      }
    }
  else 
    {
      if (!s->val) {
	fatal("pointer '%s' not initialized", s->name);
	return 1;
      }
    }

  s->hidden &= ~64;
  return 1;
}


static Element *
eval_funcall(Element *e, Lextok *now)
{
  Lextok *arg;
  Lextok *f;
  Lextok *par;
  Value *values = (Value *)0;
  int *types = (int *)0;
  int nargs, k;
  char *sp;

  /* Push the return address */
  sp = stack_push(X->stk, sizeof(Element *));
  *(Element **) sp = e ? e->nxt : ZE;
  
  /* Push the return symbol */
  sp = stack_push(X->stk, sizeof(Lextok *));
  *(Lextok **) sp = now->lft;

  Expand_Ok ++;

  /* Evaluate the arguments */
  for (arg = now->rgt->lft, nargs = 0; arg; arg = arg->rgt, nargs ++)
    {
      values = (Value *) erealloc(values, (nargs + 1) * sizeof(Value));
      types = (int *) erealloc(types, (nargs + 1) * sizeof(int));
      eval_expr(ZE, arg->lft, &values[nargs]);
      types[nargs] = node_type(arg->lft);      
    }

  Expand_Ok --;

  if (now->rgt->ntyp == FNAME) /* normal function call */
    {
      Defun *fun;

      if (!(fun = getdefun(now->rgt->sym)))
	fatal("defun not found (funcall)", 0);

      /* Enter a new run-time scope */
      rt_enter(X, fun->body, FUNCALL_TAG);

      /* Set parameters to actual values */
      for (f = fun->prms, k = 0; f; f = f->rgt)
	for (par = f->lft; par; par = par->rgt, k ++)
	  {
	    if (k >= nargs)
	      fatal("too few arguments for '%s()'", fun->fun->name);
	      
	    if (types[k] == STRUCT)
	      fatal("struct argument for '%s()'", fun->fun->name);

	    assign(par, &values[k], types[k]);
	  }
    } 
  else /* function pointer call */
    {
      Symbol *fun;
      Symbol *s;
      int n;

      fun = struct_sym(now->rgt->rgt, now->rgt->rgt->sym);
      n = eval(struct_tok(now->rgt->rgt, now->rgt->rgt->sym)->lft);

      if (!fun->type)
	fatal("undeclared variable '%s'", fun->name);

      if (!fun->Sval || !fun->Sval[n + 1])
	fatal("null function pointer call '%s'", fun->name);
      
      rt_enter(X, (Sequence *) fun->Sval[n + 1], FUNCALL_TAG);

      par = (Lextok *) emalloc(sizeof(Lextok));
      for (s = X->rt_seq->this->symtab, k = 0; s; s = s->next)
	{
	  /* is it a formal parameter ? */
	  if (s->hidden & 16)
	    {
	      if (k >= nargs)
		fatal("too few arguments for '%s()'", fun->name);
	      
	      if (types[k] == STRUCT)
		fatal("struct argument for '%s()'", fun->name);
	      
	      par->sym = s;
	      assign(par, &values[k], types[k]);
	      k ++;
	    }
	}      
      free(par);
    }

  /* Initialize local variables */
  initlocals(X->rt_seq->this);
 
  return X->rt_seq->this->frst;
}


static Element *
eval_funpoll(Element *e, Lextok *now)
{
  Defun *fun;
  Value *values = (Value *)0;
  int *types = (int *)0;
  int nargs;
  Lextok *arg, *f, *par;
  int k;

  fun = getdefun(now->rgt->sym);
  
  if (!fun)
    fatal("defun not found (funcall)", 0);

  /* Evaluate the arguments */
  for (arg = now->rgt->lft, nargs = 0; arg; arg = arg->rgt, nargs ++)
    {
      values = (Value *) erealloc(values, (nargs + 1) * sizeof(Value));
      types  = (int *) erealloc(types, (nargs + 1) * sizeof(int));
      eval_expr(ZE, arg->lft, &values[nargs]);
      types [nargs] = node_type(arg->lft);      
    }

  rt_exit(X, FUNCALL_TAG);
  rt_enter(X, fun->body, FUNCALL_TAG);

  /* Set parameters to actual values */
  for (f = fun->prms, k = 0; f; f = f->rgt)
    for (par = f->lft; par; par = par->rgt, k ++)
    {
      int t;

      if (k >= nargs)
	fatal("too few arguments for '%s()'\n", fun->fun->name);

      assign(par, &values[k], types[k]);
    }
  
  /* Initialize local variables */
  initlocals(X->rt_seq->this);

  return fun->body->frst;  
}


static int
nonprogress(void)	/* np_ */
{	RunList	*r;
	Symbol	*z;
	for (r = run; r; r = r->nxt)
	{	z = has_lab(r->pc);
		if (z && strncmp(z->name, "progress", 8))
			return 0;
	}
	return 1;
}


Element *
eval_expr(Element *e, Lextok *now, Value *vp)
{
  Value v1;
  Value v2;
  Value v3;

  if (now) 
    {
      lineno = now->ln;
      Fname  = now->fn;
#if (DEBUG >= 1)
      printf("\teval expression ");
      comment(stdout, now, 0);
      printf("\n");
#endif
      switch (now->ntyp) 
	{
	case CONST: 
	  vp->value = now->val;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;		
	  
	case   '!': 
	  eval_expr(e, now->lft, &v1);
	  vp->value = !v1.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case  UMIN: 
	  eval_expr(e, now->lft, &v1);
	  vp->value = - v1.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '~': 
	  eval_expr(e, now->lft, &v1);
	  vp->value = ~ v1.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '/': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value / v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '*': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value * v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;
	  
	case   '-': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value - v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '+': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value + v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '%': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value % v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    LT: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value < v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    GT: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value > v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '&': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value & v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '^': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value ^ v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '|': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value | v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    LE: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value <= v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    GE: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value >= v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    NE: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);

	  if (node_type(now->lft) & POINTER
	      && node_type(now->rgt) & POINTER)
	    vp->value = v1.lval != v2.lval;
	  else
	    vp->value = v1.value != v2.value;

	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    EQ: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);

	  if (node_type(now->lft) & POINTER
	      && node_type(now->rgt) & POINTER)
	    vp->value = v1.lval == v2.lval;
	  else
	    vp->value = v1.value == v2.value;

	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    OR: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value || v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   AND: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value && v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case LSHIFT: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value << v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case RSHIFT: 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt, &v2);
	  vp->value = v1.value >> v2.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   '?': 
	  eval_expr(e, now->lft, &v1);
	  eval_expr(e, now->rgt->lft, &v2);
	  eval_expr(e, now->rgt->rgt, &v3);
	  vp->value = v1.value ? v2.value : v3.value;
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case     'p': 
	  vp->value = remotevar(now); /* only _p allowed */
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case     'q': 
	  vp->value = remotelab(now);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case     'R': 
	  vp->value = qrecv(now, 0); /* test only    */
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case     LEN: 
	  vp->value = qlen(now);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    FULL: 
	  vp->value = qfull(now);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   EMPTY: 
	  vp->value = (qlen(now)==0);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case   NFULL: 
	  vp->value = (!qfull(now));
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case  NEMPTY: 
	  vp->value = (qlen(now)>0);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case ENABLED: 
	  if (s_trail) 
	    vp->value = 1;
	  else
	    vp->value = pc_enabled(now->lft);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case    EVAL: 
	  eval_expr(e, now->lft, vp);
	  break;

	case  PC_VAL: 
	  vp->value = pc_value(now->lft);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	case NONPROGRESS: 
	  vp->value = nonprogress();
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;
	  
	case  NAME:
	  vp->value = getval(now);
	  vp->lval  = getlval(now);
	  vp->heap  = is_on_heap(now); 
	  break;

	case FNAME:
	  vp->value = get_func_address(now);
	  vp->lval  = 0;
	  vp->heap  = 0;
	  break;

	  /* CHANGE: dynamic memory stuff */
	case LVAL: 
	  vp->lval = getlval(now->lft);
	  vp->heap = 0;
	  break;

	case __NULL: 
	  vp->lval  = 0;
	  vp->value = 0;
	  vp->heap  = 0;
	  break;

	case NEW:
	  v1.lval = eval_new(now->rgt);
	  v1.heap = 1;
	  assign(now->lft, &v1, POINTER | now->rgt->sym->type);
	  vp->value = 1; /* allways executable */	 
	  return e->nxt;

	case DELETE: 
	  eval_delete(now);
	  vp->value = 1; /* allways executable */
	  return e->nxt;

	  /* CHANGE: function call-return */
	case FUNCALL: 
	  vp->value = 1; /* allways executable */
	  return eval_funcall(e, now);

	case RETFUN:
	  vp->value = 1; /* allways executable */
	  return eval_funpoll(e, now);

	case RETURN:  
	  vp->value = 1; /* allways executable */
	  eval_expr(ZE, now->lft, &v1);
	  return eval_funret(&v1, node_type(now->lft));

	case TIMEOUT: 
	  vp->value = Tval;
	  return e->nxt;

	case RUN: 
	  vp->value = TstOnly ? 1 : enable(now);
	  return e->nxt;
	  
	case   's': 
	  vp->value = qsend(now);       /* send         */
	  return e->nxt;
	  
	case   'r': 
	  vp->value = qrecv(now, 1);	/* receive or poll */
	  return e->nxt;

	case   'c': 
	  eval_expr(e, now->lft, vp);	/* condition    */
	  return e->nxt;

	case PRINT: 
	  vp->value = TstOnly ? 1 : interprint(now);
	  return e->nxt;

	case  ASGN:
	  eval_expr(e, now->rgt, &v1);
	  assign(now->lft, &v1, node_type(now->rgt));	  
	  vp->value = 1; /* allways executable */	    
	  return e->nxt;
	
	case ASSERT: 
	  eval_expr(e, now->lft, &v1);

	  if (TstOnly || v1.value)
	    vp->value = 1;
	  else 
	    {
	      non_fatal("assertion violated", (char *) 0);
	      wrapup(1); /* doesn't return */
	    }

	  vp->lval = 0;
	  vp->heap = 0;
	  return e->nxt;

	case  IF: case DO: case BREAK: case UNLESS:	/* compound */
	case   '.': 
	  vp->value = 1;   /* return label for compound */
	  return e->nxt;	

	case   STOP:
	  vp->value = 0;   /* stop state */
	  return ZE;
  
	case  ELSE: 
	  vp->value = 1; /* hit here in guided trails */
	  return e->nxt; 

	default   : printf("spin: bad node type %d (run)\n", now->ntyp);
		    fatal("aborting", 0);
	}
      
      return ZE;  
    }
  
  vp->value = 0;
  vp->lval = 0;
  vp->heap = 0;

  return ZE;
}

/* A front end for simple expression evaluation */
int
eval(Lextok *now)
{
  Value v;

  eval_expr(ZE, now, &v);

  return v.value;
}

static int
interprint(Lextok *n)
{	Lextok *tmp = n->lft;
	char c, *s = n->sym->name;
	int i, j; char lbuf[16];
	char buf[1024];

	buf[0] = '\0';
	if (!no_print)
	if (!s_trail || depth >= jumpsteps) {
	for (i = 0; i < (int) strlen(s); i++)
		switch (s[i]) {
		case '\"': break; /* ignore */
		case '\\':
			 switch(s[++i]) {
			 case 't': strcat(buf, "\t"); break;
			 case 'n': strcat(buf, "\n"); break;
			 default:  goto onechar;
			 }
			 break;
		case  '%':
			 if ((c = s[++i]) == '%')
			 {	strcat(buf, "%"); /* literal */
				break;
			 }
			 if (!tmp)
			 {	non_fatal("too few print args %s", s);
				break;
			 }
			 j = eval(tmp->lft);
			 tmp = tmp->rgt;
			 switch(c) {
			 case 'c': sprintf(lbuf, "%c", j); break;
			 case 'd': sprintf(lbuf, "%d", j); break;
			 case 'o': sprintf(lbuf, "%o", j); break;
			 case 'u': sprintf(lbuf, "%u", (unsigned) j); break;
			 case 'x': sprintf(lbuf, "%x", j); break;
			 default:  non_fatal("bad print cmd: '%s'", &s[i-1]);
				   lbuf[0] = '\0'; break;
			 }
			 goto append;
		default:
onechar:		 lbuf[0] = s[i]; lbuf[1] = '\0';
append:			 strcat(buf, lbuf);
			 break;
		}
		dotag(buf);
	}
	return 1;
}

static int
Enabled1(Lextok *n)
{	int i; int v = verbose;

	if (n)
	switch (n->ntyp) {
	case 'c':
		if (has_typ(n->lft, RUN))
			return 1;	/* conservative */
		/* else fall through */
	default:	/* side-effect free */
		verbose = 0;
E_Check++;
                i = eval(n);
E_Check--;
		verbose = v;
		return i;

	case PRINT:  case  ASGN:
	case ASSERT:
		return 1;

	case 's':
		if (q_is_sync(n))
		{	if (Rvous) return 0;
			TstOnly = 1; verbose = 0;
E_Check++;
                        i = eval(n);
E_Check--;
			TstOnly = 0; verbose = v;
			return i;
		}
		return (!qfull(n));
	case 'r':
		if (q_is_sync(n))
			return 0;	/* it's never a user-choice */
		n->ntyp = 'R'; verbose = 0;
E_Check++;
                i = eval(n);
E_Check--;
		n->ntyp = 'r'; verbose = v;
		return i;
	}
	return 0;
}

int
Enabled0(Element *e)
{	SeqList *z;

	if (!e || !e->n)
		return 0;

	switch (e->n->ntyp) {
	case STOP:
		return X->pid == (nproc-nstop-1);
	case '.':
		return 1;
	case GOTO:
		if (Rvous) return 0;
		return 1;
	case UNLESS:
		return Enabled0(e->sub->this->frst);
	case ATOMIC:
	case D_STEP:
	case NON_ATOMIC:
		return Enabled0(e->n->sl->this->frst);
	}
	if (e->sub)	/* true for IF, DO, and UNLESS */
	{	for (z = e->sub; z; z = z->nxt)
			if (Enabled0(z->this->frst))
				return 1;
		return 0;
	}
	for (z = e->esc; z; z = z->nxt)
	{	if (Enabled0(z->this->frst))
			return 1;
	}
#if 0
	printf("enabled1 ");
	comment(stdout, e->n, 0);
	printf(" ==> %s\n", Enabled1(e->n)?"yes":"nope");
#endif
	return Enabled1(e->n);
}

int
pc_enabled(Lextok *n)
{	int i = nproc - nstop;
	int pid;
	int result = 0;
	RunList *Y, *oX;

	pid = eval(n);

	for (Y = run; Y; Y = Y->nxt)
		if (--i == pid)
		{	oX = X; X = Y;
			result = Enabled0(Y->pc);
			X = oX;
			break;
		}
	return result;
}
